"""
This module is used to make HTTP requests to the parents endpoint.
It can only be used by parents, and not by students.
"""

from typing import Optional
from .core import Module
from ..types import Response, Date
from ..utils import convert_date


class ParentsTalks:
    """
    This class provides endpoints to interact with the parents' talks API.
    """

    def __init__(self, module: "ParentsModule"):
        self.module = module

    async def teachers(self, student_id: int) -> Response:
        """
        Get the list of teachers for a specific student.

        :param student_id: The ID of the student.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request("GET", f"/{student_id}/talks/teachers")

    async def frames(
        self,
        student_id: int,
        teacher_id: int,
        start: Optional[Date] = None,
        end: Optional[Date] = None,
    ) -> Response:
        """
        Get the frames for a specific student and teacher within a given time range.

        :param student_id: The ID of the student.
        :param teacher_id: The ID of the teacher.
        :param start: Optional. The start date of the time range.
        :param end: Optional. The end date of the time range.
        :return: The response from the Classeviva API.
        :rtype: dict
        :raises ValueError: If the end date is earlier than the start date.
        """
        if start and end and end < start:
            raise ValueError("end must be greater than start")

        start = convert_date(start) if start else ""
        end = convert_date(end) if start else ""

        return await self.module.request(
            "GET",
            f"/{student_id}/talks/getframes/{teacher_id}/{start}/{end}".rstrip("/"),
        )

    async def all(
        self,
        student_id: int,
        start: Optional[Date] = None,
        end: Optional[Date] = None,
    ) -> Response:
        """
        Get all the frames for a specific student within a given time range.

        :param student_id: The ID of the student.
        :param start: Optional. The start date of the time range.
        :param end: Optional. The end date of the time range.
        :return: The response from the Classeviva API.
        :rtype: dict
        :raises ValueError: If the end date is earlier than the start date.
        """
        if start and end and end < start:
            raise ValueError("end must be greater than start")

        start = convert_date(start) if start else ""
        end = convert_date(end) if start else ""

        return await self.module.request(
            "GET", f"/{student_id}/talks/teachersframes2/{start}/{end}".rstrip("/")
        )

    async def book(
        self,
        student_id: int,
        teacher_id: int,
        frame_id: int,
        slot_bitmask: int,
        *,
        cellphone: Optional[str] = None,
        email: Optional[str] = None,
        other: Optional[str] = None,
    ) -> Response:
        """
        Book a frame for a specific student with the given details.

        :param student_id: The ID of the student.
        :param teacher_id: The ID of the teacher.
        :param frame_id: The ID of the frame.
        :param slot_bitmask: The bitmask for the time slots.
        :param cellphone: Optional. The cellphone number.
        :param email: Optional. The email address.
        :param other: Optinoal. Other details.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "POST",
            f"/{student_id}/talks/book/{teacher_id}/{frame_id}/{slot_bitmask}",
            json={"cell": cellphone, "email": email, "altro": other},
        )

    async def booked(
        self,
        student_id: int,
        start: Optional[Date] = None,
        end: Optional[Date] = None,
    ) -> Response:
        """
        Get the booked frames for a specific student within a given time range.

        :param student_id: The ID of the student.
        :param start: Optional. The start date of the time range.
        :param end: Optional. The end date of the time range.
        :return: The response from the Classeviva API.
        :rtype: dict
        :raises ValueError: If the end date is earlier than the start date.
        """
        if start and end and end < start:
            raise ValueError("end must be greater than start")

        start = convert_date(start) if start else ""
        end = convert_date(end) if start else ""

        return await self.module.request(
            "GET", f"/{student_id}/talks/mytalks/{start}/{end}".rstrip("/")
        )

    async def delete(self, student_id: int, talk_id: int) -> Response:
        """
        Delete a specific talk for a student.

        :param student_id: The ID of the student.
        :param talk_id: The ID of the talk.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "POST", f"/{student_id}/talks/delete/{talk_id}"
        )

    async def message(self, student_id: int, talk_id: int, message: str) -> Response:
        """
        Send a message for a specific talk.

        :param student_id: The ID of the student.
        :param talk_id: The ID of the talk.
        :param message: The message to send.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "POST", f"/{student_id}/talks/mymessage/{talk_id}", json={"myMsg": message}
        )

    async def read(self, student_id: int, talk_id: int, read: bool) -> Response:
        """
        Mark a specific talk as read or unread.

        :param student_id: The ID of the student.
        :param talk_id: The ID of the talk.
        :param read: Whether to mark the talk as read or unread.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "POST",
            f"/{student_id}/talks/teachermessage/{talk_id}",
            json={"messageRead": read},
        )


class ParentsOverallTalks:
    """
    Represents a class for managing overall talks for parents.
    """

    def __init__(self, module: "ParentsModule"):
        self.module = module

    async def list(self, student_id: int) -> Response:
        """
        Get the list of overall talks for a specific student.

        :param student_id: The ID of the student.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request("GET", f"/{student_id}/overalltalks/list")

    async def frames(
        self, student_id: int, overalltalk_id: int, teacher_id: int
    ) -> Response:
        """
        Get the frames for a specific overall talk.

        :param student_id: The ID of the student.
        :param overalltalk_id: The ID of the overall talk.
        :param teacher_id: The ID of the teacher.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "GET", f"/{student_id}/overalltalks/getframes/{overalltalk_id}/{teacher_id}"
        )

    async def book(
        self,
        student_id: int,
        overalltalk_id: int,
        teacher_id: int,
        frame_id: int,
        slot_number: int,
    ) -> Response:
        """
        Book a slot for a specific overall talk.

        :param student_id: The ID of the student.
        :param overalltalk_id: The ID of the overall talk.
        :param teacher_id: The ID of the teacher.
        :param frame_id: The ID of the frame.
        :param slot_number: The slot number.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "POST",
            f"/{student_id}/overalltalks/book/{overalltalk_id}"
            f"/{teacher_id}/{frame_id}/{slot_number}",
        )

    async def delete(
        self, student_id: int, overalltalk_id: int, event_id: int
    ) -> Response:
        """
        Delete a specific overall talk event.

        :param student_id: The ID of the student.
        :param overalltalk_id: The ID of the overall talk.
        :param event_id: The ID of the event.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "DELETE", f"/{student_id}/overalltalks/delete/{overalltalk_id}/{event_id}"
        )


class ParentsPayments:
    """
    This class provides endpoints for handling parents' payments.
    """

    def __init__(self, module: "ParentsModule"):
        self.module = module

    async def payments(self, student_id: int) -> Response:
        """
        Get the payments for a specific student.

        :param student_id: The ID of the student.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request("GET", f"/{student_id}/pagoonline/payments")

    async def download_file(self, student_id: int, attach_id: int) -> Response:
        """
        Download a file associated with a payment.

        :param student_id: The ID of the student.
        :param attach_id: The ID of the attachment.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "GET", f"/{student_id}/pagoonline/downloadfile/{attach_id}"
        )

    async def privacy(self, student_id: int) -> Response:
        """
        Get the privacy settings for a specific student.

        :param student_id: The ID of the student.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request("GET", f"/{student_id}/pagoonline/getprivacy")

    async def download_file_privacy(self, student_id: int, file_id: int) -> Response:
        """
        Download a privacy file associated with a student.

        :param student_id: The ID of the student.
        :param file_id: The ID of the privacy file.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "POST", f"/{student_id}/pagoonline/downloadfileprivacy/{file_id}"
        )

    async def set_privacy(
        self,
        student_id: int,
        name: str,
        surname: str,
        fiscal_code: str,
        relationship: str,
        iban: str,
        flag_privacy: bool,
        flag_coord: bool,
    ) -> Response:
        """
        Set the privacy settings for a specific student.

        :param student_id: The ID of the student.
        :param name: The name of the parent.
        :param surname: The surname of the parent.
        :param fiscal_code: The fiscal code of the parent.
        :param relationship: The relationship of the parent to the student.
        :param iban: The IBAN of the parent.
        :param flag_privacy: The privacy flag.
        :param flag_coord: The coordination flag.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.module.request(
            "POST",
            f"/{student_id}/pagoonline/setprivacy",
            json={
                "name": name,
                "surname": surname,
                "fiscalCode": fiscal_code,
                "relationship": relationship,
                "iban": iban,
                "flagPrivacy": flag_privacy,
                "flagCoord": flag_coord,
            },
        )


class ParentsModule(Module):
    """
    This module is used to make HTTP requests to Classeviva's parents module.
    """

    endpoint = "parents"

    @property
    def talks(self) -> ParentsTalks:
        """
        Get talks-related endpoints, which have been separated
        from here to avoid making too many functions in one class
        for the same endpoint.
        """
        return ParentsTalks(self)

    @property
    def overall_talks(self) -> ParentsOverallTalks:
        """
        Get overall talks-related endpoints, which have been separated
        from here to avoid making too many functions in one class
        for the same endpoint.
        """
        return ParentsOverallTalks(self)

    @property
    def payments(self) -> ParentsPayments:
        """
        Get payments-related endpoints, which have been separated
        from here to avoid making too many functions in one class
        for the same endpoint.
        """
        return ParentsPayments(self)



"""
This module is used to make HTTP requests to the students endpoint.
It can be used by students and parents, and not by teachers.
"""

import os
from datetime import datetime, date
from typing import Optional, Any, IO
from urllib.parse import urljoin
from io import BytesIO, StringIO
from base64 import b64encode
from .core import BaseModule
from ..enums import EventCode, NoteType
from ..types import Date, Response
from ..utils import convert_date


class StudentHomeworks:
    """
    Represents a collection of methods for managing student homeworks.
    """

    def __init__(self, module: "StudentsModule"):
        self.module = module

    async def all(self, student_id: int) -> Response:
        """
        Retrieves all of the student's homeworks.

        :param student_id: The ID of the student.
        :return: The response from the Classeviva API.
        :rtype: dict
        """

        return await self.module.request("GET", f"/{student_id}/homeworks")

    async def download_teacher_file(
        self, student_id: int, event_code: str, file_id: int
    ) -> Response:
        """
        Downloads a teacher file associated with a specific homework.

        :param student_id: The ID of the student.
        :param event_code: The event code of the homework.
        :param file_id: The ID of the file.
        :return: The response from the Classeviva API.
        :rtype: dict
        """

        return await self.module.request(
            "GET",
            f"/{student_id}/homeworks/downloadTeacherFile/{event_code}/{file_id}",
        )

    async def insert_student_msg(
        self, student_id: int, event_code: str, homework_id: int, message: str
    ) -> Response:
        """
        Inserts a message from the student for a specific homework.

        :param student_id: The ID of the student.
        :param event_code: The event code of the homework.
        :param homework_id: The ID of the homework.
        :param message: The message to insert.
        :return: The response from the Classeviva API.
        :rtype: dict
        """

        return await self.module.request(
            "POST",
            f"/{student_id}/homeworks/insertStudentMsg/{event_code}/{homework_id}",
            json={"studentMsg": message},
        )

    async def upload_student_file(
        self,
        student_id: int,
        event_code: str,
        homework_id: int,
        file: IO[Any],
        filename: Optional[str] = None,
    ) -> Response:
        """
        Uploads a file from the student for a specific homework.

        :param student_id: The ID of the student.
        :param event_code: The event code of the homework.
        :param homework_id: The ID of the homework.
        :param file: The file to upload.
        :param filename: The name of the file (optional).
        :return: The response from the Classeviva API.
        :rtype: dict
        """

        payload = {}
        if isinstance(file, (BytesIO, StringIO)):
            cont = file.getvalue()
        else:
            cont = file.read()

        if not filename and not file.name:
            raise ValueError("a file name must be specified")

        payload["file"] = await self.module.client.loop.run_in_executor(
            None, lambda: b64encode(cont).decode("utf-8")
        )
        payload["filename"] = os.path.basename(getattr(file, "name", None) or filename)
        return await self.module.request(
            "POST",
            f"/{student_id}/homeworks/uploadStudentFile/{event_code}/{homework_id}",
            json=payload,
        )

    async def set_teacher_msg_status(
        self,
        student_id: int,
        event_code: str,
        homework_id: int,
        read: bool = True,
    ) -> Response:
        """
        Sets the status of a teacher message for a specific homework.

        :param student_id: The ID of the student.
        :param event_code: The event code of the homework.
        :param homework_id: The ID of the homework.
        :param read: The status of the message (default: True).
        :return: The response from the Classeviva API.
        :rtype: dict
        """

        return await self.module.request(
            "POST",
            f"/{student_id}/homeworks/setTeacherMsgStatus/{event_code}/{homework_id}",
            json={"messageRead": read},
        )

    async def remove_student_file(
        self, student_id: int, event_code: str, homework_id: int, file_id: int
    ) -> Response:
        """
        Removes a file uploaded by the student for a specific homework.

        :param student_id: The ID of the student.
        :param event_code: The event code of the homework.
        :param homework_id: The ID of the homework.
        :param file_id: The ID of the file.
        :return: The response from the Classeviva API.
        :rtype: dict
        """

        return await self.module.request(
            "POST",
            f"/{student_id}/homeworks/removeStudentFile/{event_code}/{homework_id}/{file_id}",
        )


class StudentsModule(BaseModule):
    """
    This module is used to make HTTP requests to Classeviva's students module.
    """

    endpoint = "students"

    @property
    def homeworks(self) -> StudentHomeworks:
        """
        Get homeworks-related endpoints, which have been separated
        from here to avoid making too many functions in one class
        for the same endpoint.

        Note that this is for media that may be attached to homeworks.
        If you are looking for the homeworks from the agenda, use the
        :meth:`agenda` or the :meth:`overview` method instead.
        """
        return StudentHomeworks(self)

    async def calendar(
        self,
        student_id: int,
        start: Optional[Date] = None,
        end: Optional[Date] = None,
    ) -> Response:
        """
        Get information about the school's working, non-working and holiday days.

        .. note::
            Without the dates, it returns the whole calendar
            from the beginning to the end of the school year.

        :param student_id: The ID of the student.
        :param start: The start date of the calendar (optional).
        :param end: The end date of the calendar (optional).
        :return: The response from the Classeviva API.
        :rtype: dict
        :raises ValueError: If the end date is before the start date.
        """
        if start or end:
            start = getattr(start, "date", lambda: start)() or date.today()
            end = getattr(end, "date", lambda: end)() or date.today()

            if end < start:
                raise ValueError("end must be greater than start")

            start = convert_date(start)
            end = convert_date(end)
            ret = await self.request("GET", f"/{student_id}/calendar/{start}/{end}")
        else:
            ret = await self.request("GET", f"/{student_id}/calendar/all")

        return ret

    async def absences(
        self,
        student_id: int,
        start: Optional[Date] = None,
        end: Optional[Date] = None,
    ) -> Response:
        """
        Get the student's absences.

        :param student_id: The ID of the student.
        :param start: Optional. The start date of the absences, or the single absence day.
        :param end: Optional. The end date of the absences.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        if start or end:
            start = start or date.today()
            start = start.date() if isinstance(start, datetime) else start

            end = end.date() if isinstance(end, datetime) else end
            end = end if end and end >= start else ""

            start = convert_date(start)
            end = convert_date(end) if end else ""
            ret = await self.request(
                "GET", f"/{student_id}/absences/details/{start}/{end}"
            )
        else:
            ret = await self.request("GET", f"/{student_id}/absences/details")

        return ret

    async def agenda(
        self,
        student_id: int,
        start: Date,
        end: Date,
        event_code: Optional[EventCode] = None,
    ) -> Response:
        """
        Get the student's agenda (as in events, homeworks, etc.).

        :param student_id: The ID of the student.
        :param start: The start date of the agenda.
        :param end: The end date of the agenda.
        :param event_code: Optional. The event code to filter the agenda by.

        :return: The response from the Classeviva API.
        :rtype: dict
        :raises ValueError: If the end date is before the start date.
        """
        start = getattr(start, "date", lambda: start)()
        end = getattr(end, "date", lambda: end)()

        if end < start:
            raise ValueError("end must be greater than start")

        start = convert_date(start)
        end = convert_date(end)
        if event_code:
            ret = await self.request(
                "GET", f"/{student_id}/agenda/{event_code}/{start}/{end}"
            )
        else:
            ret = await self.request("GET", f"/{student_id}/agenda/all/{start}/{end}")

        return ret

    async def lessons(
        self,
        student_id,
        start: Date,
        end: Optional[Date] = None,
        *,
        subject: Optional[int] = None,
    ) -> Response:
        """
        Get the student's lessons.

        :param student_id: The ID of the student.
        :param start: The date from which to start retrieving lessons,
                            or the day to get lessons of.
        :param end: Optional. The date until which to retrieve lessons.
        :param subject: Optional. The ID of the subject to get the lessons of.

        :return: The response from the Classeviva API.
        :rtype: dict
        """

        today = not (start and end)

        start = convert_date(start, today=today)
        end = convert_date(end, today=today) if end else None

        base = f"/{student_id}/lessons"
        url = f"{base}-status/"
        params = [start, end] if not today else [start]
        if subject:
            if today:
                params.append(start)
            params.append(subject)

        join = "/".join(str(p) for p in params)
        url = urljoin(url, join)
        ret = await self.request("GET", url, raise_for_status=False)

        # Handle 404s by using the endpoint without status,
        # because the API limits the range of dates for status requests.
        # With this, we can still return lessons but without statuses.
        if ret["status"] == 404:
            ret = await self.request("GET", urljoin(base + "/", join))

        return ret

    async def periods(self, student_id: int) -> Response:
        """
        Get the student's school year periods.

        :param student_id: The ID of the student.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.request("GET", f"/{student_id}/periods")

    async def subjects(self, student_id: int) -> Response:
        """
        Get the student's subjects.

        :param student_id: The ID of the student.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.request("GET", f"/{student_id}/subjects")

    async def grades(self, student_id: int, subject: Optional[int] = None) -> Response:
        """
        Get the student's current grades.

        :param student_id: The ID of the student.
        :param subject: Optional. The ID of the subject to get the grades of.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        if subject:
            return await self.request(
                "GET", f"/{student_id}/grades2/subjects/{subject}"
            )

        return await self.request("GET", f"/{student_id}/grades2")

    async def notes(
        self,
        student_id: int,
        type: Optional[NoteType] = None,  # pylint: disable=redefined-builtin
        event: Optional[int] = None,
    ) -> Response:
        """
        Get the student's disciplinary notes.

        :param student_id: The ID of the student.
        :param type: Optional. The type of note to get.
        :param event: Optional. The ID of the note to get.

        :return: The response from the Classeviva API.
        :rtype: dict
        :raises ValueError: If the note ID is provided without the note type.
        """
        if event and not type:
            raise ValueError("event requires type")

        url = f"/{student_id}/notes/"
        meth = "GET"
        if type:
            url = urljoin(url, type.value) + "/"
            if event:
                url = urljoin(url, f"read/{event}")
                meth = "POST"
        else:
            url = urljoin(url, "all")

        return await self.request(meth, url)

    async def didactics(
        self, student_id: int, content_id: Optional[int] = None
    ) -> Response:
        """
        Get the didactics for a student.

        :param student_id: The ID of the student.
        :param content_id: Optional. The content ID.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        url = f"/{student_id}/didactics"
        if content_id:
            url = urljoin(url, f"item/{content_id}")

        return await self.request("GET", url)

    async def documents(
        self,
        student_id: int,
        hash: Optional[str] = None,  # pylint: disable=redefined-builtin
        *,
        check: Optional[bool] = False,
    ) -> Response:
        """
        Get the documents for a student.

        :param student_id: The ID of the student.
        :param hash: Optional. The hash of the document.
        :param check: Optional. Whether to check the document.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        url = f"/{student_id}/documents/"
        if hash:
            url = urljoin(url, f"{'check' if check else 'read'}/{hash}")

        return await self.request("POST", url)

    async def schoolbooks(self, student_id: int) -> Response:
        """
        Get the student's schoolbooks.

        :param student_id: The ID of the student.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.request("GET", f"/{student_id}/schoolbooks")

    async def register_config(self, student_id: int) -> Response:
        """
        Get the register configuration.

        :param student_id: The ID of the student.

        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.request("GET", f"/{student_id}/register-config")

    async def overview(
        self,
        student_id: int,
        start: Date,
        end: Optional[Date] = None,
    ) -> Response:
        """
        Get the student's agenda, lessons, events, grades and notes of a period, all in one request.

        :param student_id: The ID of the student.
        :param start: The start date of the overview or
                           the single day to get the overview of.
        :param end: Optional. The end date of the overview.

        :return: The response from the Classeviva API.
        :rtype: dict
        :raises ValueError: If the end date is before the start date.
        """
        if end and end < start:
            raise ValueError("end must be greater than start")

        start = convert_date(start)
        end = convert_date(end) if end else ""
        return await self.request("GET", f"/{student_id}/overview/all/{start}/{end}")

    async def virtual_classes(self, student_id: int):
        """
        Get the student's virtual classes.

        :param student_id: The ID of the student.
        :return: The response from the Classeviva API.
        :rtype: dict
        """
        return await self.request("GET", f"/{student_id}/virtualclasses/myclasses")



"""
This helper contains the Calendar and Period classes, which are used to
interact with the calendar endpoint of the Classeviva API.

The Calendar class is used to get information about the user's calendar,
such as school days, absences, events, grades, notes, and more.

The Period class is used to get information about the school
periods (e.g. first quarter, second quarter, etc.) and
everything that happened during them.
"""

from .core import Calendar
from .period import Period



"""
This is responsible for handling the authentication of the client.
It is used internally by the client and should not be used directly.
"""

from urllib.parse import urljoin
from datetime import datetime, timezone
from typing import Optional

import bcrypt

from aiohttp import ClientSession, ClientResponseError
from .errors import AuthenticationError, MultiIdentFound
from .modules.core import Module, Response
from .core import CLIENT_USER_AGENT, CLIENT_DEV_APIKEY, CLIENT_CONTENT_TP
from .utils import find_exc


class AuthenticationModule(Module):
    """
    This module is responsible for handling the authentication of the client.
    """

    endpoint = "auth"

    async def login(
        self, username: str, password: str, identity: Optional[str] = None
    ) -> dict:
        """
        Get a temporary token to use as authentication with the given username and password.

        It works by preparing caches for all the future requests and
        then sending a POST request to the Classeviva API to get a
        temporary token to use in requests.

        If a token is found cached and unexpired, that one will be used instead
        of requesting a new one. Otherwise, this function will request a new
        one, and cache it again for future requests.

        :param username: The user's username, or email or badge to authenticate with.
        :param password: The user's password.
        :param identity: The user's identity, in case multiple are found.
        :return: The direct response from the Classeviva API containing the token.
        """
        with self.get_cache() as cache:
            if "salt" not in cache:
                cache["salt"] = bcrypt.gensalt()

            # hash the password to cache it
            hashed_pw = bcrypt.hashpw(password.encode(), cache["salt"]).decode()
            if "logins" not in cache:
                cache["logins"] = {}

            cache_key = f"{username}:{hashed_pw}"
            login_cache = cache["logins"]
            if identity:
                cache_key += f":{identity}"

            # check in the cache for the token and its expiration
            if cache_key in login_cache:
                this = login_cache[cache_key]
                expires_at = datetime.fromisoformat(this["expire"])
                if expires_at > datetime.now(timezone.utc):
                    return this

            req = {"uid": username, "pass": password}
            if identity:
                req["ident"] = identity

            # do the actual request to get the token, if expired or not found
            async with ClientSession(loop=self.client.loop) as session:
                async with session.post(
                    urljoin(self.client.BASE_URL, "auth/login"),
                    headers={
                        "User-Agent": CLIENT_USER_AGENT,
                        "Z-Dev-Apikey": CLIENT_DEV_APIKEY,
                        "Content-Type": CLIENT_CONTENT_TP,
                    },
                    json=req,
                ) as resp:
                    content = await resp.json()
                    if resp.status == 422:
                        msg = {
                            "content": content,
                            "status": resp.status,
                            "status_reason": resp.reason,
                        }
                        raise find_exc(msg, AuthenticationError)

                    if "choices" in content and (
                        not identity
                        or identity not in [c["ident"] for c in content["choices"]]
                    ):
                        choices = " * " + "\n * ".join(
                            f"{c['ident']} ({c['name']})" for c in content["choices"]
                        )
                        msg = "Multiple identities have been found, but none has been specified"
                        if identity:
                            msg = "Could not find the requested identity"

                        raise MultiIdentFound(
                            f"{msg}. Possible choices are:\n{choices}"
                        )

                    try:
                        resp.raise_for_status()
                    except ClientResponseError as e:
                        raise AuthenticationError(content) from e

                    # cache response, will be re-cached as soon as the token expires
                    login_cache[cache_key] = content
                    cache["logins"] = login_cache
                    return content

    async def status(self, token: str) -> dict:
        """
        Get the status of the given token.
        This function is not used, but it is still available.

        :param token: The token to check the status of.
        :return: The direct response from the Classeviva API.
        """
        with self.get_cache() as cache:
            if "logins_status" not in cache:
                cache["logins_status"] = {}

            status = cache["logins_status"]
            if token in status:
                this = status[token]
                expires_at = datetime.fromisoformat(this["expire"])
                if expires_at > datetime.now(timezone.utc):
                    return this

            async with ClientSession(loop=self.client.loop) as session:
                async with session.get(
                    urljoin(self.client.BASE_URL, "auth/status"),
                    headers={
                        "User-Agent": CLIENT_USER_AGENT,
                        "Z-Dev-Apikey": CLIENT_DEV_APIKEY,
                        "Content-Type": CLIENT_CONTENT_TP,
                        "Z-Auth-Token": token,
                    },
                ) as resp:
                    resp.raise_for_status()
                    status[token] = (await resp.json())["status"]
                    cache["logins_status"] = status
                    return status[token]



"""
Type hints that are used across the whole project.
"""

from typing import TypedDict, Any, Dict, Union, TypeVar
from datetime import date, datetime
from .errors import ClassevivaError


class OKResponse(TypedDict):
    """
    Type hint for a successful response.
    """

    created_at: int
    content: Any
    headers: Dict[str, str]
    status: int
    status_reason: str


class ErrorResponseContent(TypedDict):
    """
    Type hint for the content of an error response.
    """

    statusCode: int
    error: str


class ErrorResponse(OKResponse):
    """
    Type hint for an error response.
    """

    content: ErrorResponseContent


Response = Union[OKResponse, ErrorResponse]

Date = Union[date, datetime]

CVVErrors = TypeVar("CVVErrors", bound=ClassevivaError)
AnyCVVError = Union[ClassevivaError, CVVErrors]  # pylint: disable=invalid-name



import os
import re
from setuptools import setup, find_packages

packages = find_packages()

with open("README.md", "r", encoding="utf-8") as f:
    long_desc = f.read()

with open(
    os.path.join(os.path.dirname(__file__), packages[0], "__init__.py"),
    "r",
    encoding="utf-8",
) as f:
    kwargs = {
        var.strip("_"): val
        for var, val in re.findall(
            r'^(__\w+__)\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
        )
    }
    kwargs["author_email"] = kwargs.pop("email", "")

setup(
    name="aiocvv",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    packages=packages,
    install_requires=["aiohttp", "appdirs", "bcrypt", "typing-extensions"],
    python_requires=">=3.7",
    **kwargs
)



"""
This module contains the custom exceptions raised by the library.
"""


class ClassevivaError(Exception):
    """
    A HTTP error occurred while communicating with the Classeviva API.
    """

    status_code = None

    def __init__(self, response):
        desc = response
        if isinstance(response, dict):
            content = response["content"]
            self.error = content["error"]
            self.message = content.get("message", "")
            self.reason = response["status_reason"]
            desc = self.error + (f": {self.message}" if self.message else "")

            if self.status_code is None or self.status_code != response["status"]:
                self.status_code = response["status"]
                desc = f"{self.status_code} {self.reason}: {desc}"

        super().__init__(desc)


# Request errors
class BadRequest(ClassevivaError):
    """
    The server returned a 4xx/500 status code.
    """

    status_code = 400


class Unauthorized(BadRequest):
    """
    The server returned a 401 status code.
    This is common when you try to make HTTP requests without an API key, or with an invalid one.
    """

    status_code = 401


class NotFound(BadRequest):
    """
    The server returned a 404 status code, meaning the requested resource doesn't exist.
    """

    status_code = 404


class UnprocessableEntity(BadRequest):
    """
    The server returned a 422 status code, meaning the request
    was well-formed but the server was unable to process it for some reason.

    This is common when you try to log in with wrong credentials.
    """

    status_code = 422


class TooManyRequests(BadRequest):
    """
    The server returned a 429 status code, meaning you've reached the rate limit.
    If you get this error, you should really wait some time before trying again.
    """

    status_code = 429


# Server errors
class InternalServerError(ClassevivaError):
    """
    The server returned a 500/5xx status code.
    """

    status_code = 500


class ServiceUnavailable(InternalServerError):
    """
    The server returned a 503 status code, meaning the requested resource is currently unavailable.
    """

    status_code = 503


# Authentication errors
class AuthenticationError(ClassevivaError):
    """
    An error occurred while trying to authenticate with the Classeviva API.
    """

    status_code = 422


class WrongCredentials(AuthenticationError):
    """
    You provided wrong credentials to the Classeviva API.
    """


class PasswordExpired(AuthenticationError):
    """
    You provided correct credentials to the Classeviva API, but your password has expired.
    This means you must renew it from the Classeviva website before you can use the API.
    """


class NoIdentAvailable(AuthenticationError):
    """
    You successfully logged in through email or nickname, but
    no valid accounts were found in the associated portfolio.
    """


class MultiIdentFound(AuthenticationError):
    """
    You successfully logged in through email or nickname, but
    multiple accounts were found in the associated portfolio.

    You should pass the specific identity to use as the `ident` parameter
    in the :class:`~aiocvv.client.ClassevivaClient` constructor.
    """



"""
This module contains the class that represents
students, teachers and parents all together.
"""

from datetime import datetime
from io import BytesIO
from typing import Optional
from .enums import UserType, GradeCode, NoteType
from .helpers import Noticeboard, Calendar
from .dataclasses import School, MIURData, Subject, Teacher as TeacherT, Grade, Note
from .utils import capitalize_name, group_by_date, parse_date


class Me:
    """
    Represents a Classeviva user, whether it's a student, a teacher or a parent.

    .. note::
        This class is not meant to be manually constructed, but to be
        used through the :class:`~aiocvv.client.ClassevivaClient` class.
    """

    def __init__(self, client, **kwargs):
        from .client import ClassevivaClient

        self.client: ClassevivaClient = client
        self.__card = kwargs
        self.__noticeboard = None

    @property
    def identity(self) -> str:
        """
        The user's identity.
        """
        return self.__card["ident"]

    @property
    def type(self) -> UserType:
        """
        The type of the logged in user.
        """
        return UserType(self.__card["usrType"])

    @property
    def id(self) -> int:
        """
        The user's ID.
        """
        return self.__card["usrId"]

    @property
    def school(self) -> School:
        """
        The user's school details.
        """
        return School(
            code=self.__card["schCode"],
            name=self.__card["schName"],
            dedication=self.__card["schDedication"],
            city=self.__card["schCity"],
            proince=self.__card["schProv"],
            miur_data=MIURData(
                self.__card["miurSchoolCode"], self.__card["miurDivisionCode"]
            ),
        )

    @property
    def first_name(self) -> str:
        """The user's first name."""
        return capitalize_name(self.__card["firstName"])

    @property
    def last_name(self) -> str:
        """The user's last name."""
        return capitalize_name(self.__card["lastName"])

    @property
    def name(self) -> str:
        """The user's full name."""
        return f"{self.first_name} {self.last_name}"

    @property
    def birth_date(self) -> datetime:
        """The user's birth date."""
        return datetime.strptime(self.__card["birthDate"], "%Y-%m-%d")

    @property
    def fiscal_code(self) -> str:
        """The user's fiscal code."""
        return self.__card["fiscalCode"]

    async def refresh(self):
        """Refresh the user's data."""
        self.__card = await getattr(self.client, self.type.name).get_card(self.id)

    async def get_enabled_apps(self):
        """Get the enabled apps for the user."""
        resp = await self.client.request("GET", "/misc/enabled-apps")
        return resp["content"].get("enabledApps", [])

    async def get_avatar(self) -> BytesIO:
        """
        Returns the user's avatar as a BytesIO object.
        """
        resp = await self.client.request("GET", f"/users/{self.identity}/avatar")
        ret = BytesIO(resp["content"])
        ret.name = f"{self.identity}.jpg"
        return ret

    @property
    def noticeboard(self) -> Noticeboard:
        """The user's noticeboard."""
        if self.__noticeboard is None:
            tp = self.type
            if tp == UserType.parent:
                tp = UserType.student

            self.__noticeboard = Noticeboard(
                self.id, getattr(self.client, tp.name + "s")
            )

        return self.__noticeboard


class Teacher(Me):
    """
    Represents a Classeviva teacher.

    .. note::
        This class is not meant to be manually constructed, but to be
        used through the :class:`~aiocvv.client.ClassevivaClient` class.

    .. warning::
        This class is not yet implemented. You have to use the
        :meth:`~aiocvv.client.ClassevivaClient.teachers.request`
        method to manually make requests to the Classeviva API.
    """

    pass


class Student(Me):
    """
    Represents a Classeviva student.

    .. note::
        This class is not meant to be manually constructed, but to be
        used through the :class:`~aiocvv.client.ClassevivaClient` class.
    """

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self.__calendar = None

    @staticmethod
    def _parse_grade(data, subjects, periods):
        return Grade(
            subject=list(filter(lambda s: s.id == data["subjectId"], subjects))[0],
            subject_code=data["subjectCode"],
            id=data["evtId"],
            code=GradeCode(data["evtCode"]),
            date=datetime.strptime(data["evtDate"], "%Y-%m-%d").date(),
            value=data["decimalValue"],
            display_value=data["displayValue"],
            position=data["displaPos"],
            family_notes=data["notesForFamily"],
            color=data["color"],
            canceled=data["canceled"],
            underlined=data["underlined"],
            period=list(filter(lambda p: p.position == data["periodPos"], periods))[0],
            component_position=data["componentPos"],
            component_description=data["componentDesc"],
            weight=data["weightFactor"],
            skill_id=data["skillId"],
            grade_master_id=data["gradeMasterId"],
            skill_description=data["skillDesc"],
            skill_code=data["skillCode"],
            skill_master_id=data["skillMasterId"],
        )

    @staticmethod
    def _parse_note(data, type_: NoteType):
        return Note(
            id=data["evtId"],
            type=type_,
            text=data["evtText"],
            date=parse_date(data.get("evtBegin", None) or data["evtDate"]),
            author_name=capitalize_name(data["authorName"]),
            read=data["readStatus"],
            end=parse_date(data["evtEnd"]) if data.get("evtEnd") else None,
        )

    @property
    def calendar(self) -> Calendar:
        """The user's calendar."""
        if self.__calendar is None:
            self.__calendar = Calendar(self.client.students, self.id)

        return self.__calendar

    async def get_subjects(self, include_grades: bool = False) -> list[Subject]:
        """
        Get the user's subjects.

        :param include_grades: Whether to include the grades for each subject.
        :return: A list of the user's subjects.
        """
        resp = await self.client.students.subjects(self.id)
        resp = resp["content"]["subjects"]
        return [
            Subject(
                teachers=[
                    TeacherT(id=t["teacherId"], name=capitalize_name(t["teacherName"]))
                    for t in subject.pop("teachers", [])
                ],
                grades=await self.get_grades(subject) if include_grades else None,
                **subject,
            )
            for subject in resp
        ]

    async def get_grades(self, subject: Optional[Subject] = None) -> list[Grade]:
        """
        Get the user's grades.

        :param subject: The subject to get the grades from.
        """
        resp = await self.client.students.grades(
            self.id, subject.id if subject else None
        )
        periods = await self.calendar.get_periods()
        subjects = await self.get_subjects()
        return [
            self._parse_grade(g, subjects, periods) for g in resp["content"]["grades"]
        ]

    async def get_notes(self, type: Optional[NoteType] = None) -> list[Note]:
        """Get the user's notes."""

        resp = await self.client.students.notes(self.id, type.value if type else None)
        resp = resp["content"]
        notes = {}
        for t in NoteType:
            notes.update(group_by_date(resp[t.value], self._parse_note, t))

        ret = []
        for notess in notes.values():
            ret += list(notess)

        return ret


class Parent(Student):
    """
    Represents a Classeviva parent, which also has access to
    the students' data (refer to :class:`~aiocvv.me.Student`).

    .. note::
        This class is not meant to be manually constructed, but to be
        used through the :class:`~aiocvv.client.ClassevivaClient` class.
    """

    pass
